#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <curses.h>
#include <string.h>

#include "config.h"

#define OXT_VERSION "0.3.1"

/** some curses compatibility (kludge!!!) */

#ifndef WA_NORMAL
  #define WA_NORMAL 0
#endif

#ifndef WA_REVERSE
  #define WA_REVERSE A_REVERSE
#endif

#ifndef WA_BOLD
  #define WA_BOLD A_BOLD
#endif

/***/

#ifdef __SONG_C
	#define global
#else
   #define global extern
#endif

/******************** TYPEDEFS FOR TABLES *****************/

typedef struct{
	int loc;
	int size;
	int type;	
	int memloc;
	int memsize;
}fielddef;

typedef struct{
	int memsize;
	int scrsize;
	int numfields;
	const fielddef*fields;
	void*dflt;
}celldef;

typedef struct{
	int celltype;
	int height,width;
	void*data;
	int patt,chan;
	char title[36];
}columndef;

typedef struct{
	int width;
	columndef*cols;
}tabledef;

/******************** FiELD CLASSES *****************/

enum { NoteFld, ByteFld, CharFld, IntFld, StringFld };
/* StringFld, IntFld, NoneFld */

/******************** CELL CLASSES ******************/

enum { Trackcell, Ordercell, Samplecell };

#define uchar unsigned char
#define uint  unsigned int

	/**/ /**/

typedef struct{
	uchar pattno;
}ordercellT;

typedef struct{
	uchar note;
	uchar ins;
	uchar vol;
	uchar fx;
	uchar fxb;
}trackcellT;

typedef struct{
	signed char*data; /* => short? */
	uint length,alloc;
	uint loopstart,loopend;
	uint freq;
	uchar volume;        /* format? */
	uchar format;
	uchar looping;
	char name[32];
}sampleT;

#ifdef __SONG_C

const fielddef trackcellFD[]={
	 {0,1,NoteFld,0,1},
	 {4,2,ByteFld,1*sizeof(char),1},
	 {7,2,ByteFld,2*sizeof(char),1},
	 {10,1,CharFld,3*sizeof(char),1},
	 {11,2,ByteFld,4*sizeof(char),1} };
const trackcellT trackcellDFLT={ 0xff, 0xff, 0xff, '0', 0};

const fielddef ordercellFD[]={{ 0,2,ByteFld,0,1 }};
const ordercellT ordercellDFLT ={ 0xff };

const fielddef samplecellFD[]={
	{0,1,CharFld, sizeof(char*)+sizeof(int)*5+sizeof(char)*1,1},
	{1,1,CharFld, sizeof(char*)+sizeof(int)*5+sizeof(char)*2,1},
	{0+3,5,IntFld,  sizeof(char*),1},                /* length*/
	{6+3,5,IntFld, sizeof(char*)+sizeof(int)*2,1},   /* loopstart*/ 
	{12+3,5,IntFld, sizeof(char*)+sizeof(int)*3,1},  /* loopend*/ 
	{18+3,5,IntFld, sizeof(char*)+sizeof(int)*4,1},  /* freq*/
	{24+3,2,ByteFld,sizeof(char*)+sizeof(int)*5,1},  /* volume*/
	{27+3,32,StringFld,sizeof(char*)+sizeof(int)*5+sizeof(char)*3,1}
};
const sampleT samplecellDFLT={NULL,0,0,0,0,0,0,'0','0',"\0"}; /* hmm */

#endif

global const celldef celltypes[]
#ifdef __SONG_C
={ 
	{sizeof(trackcellT),14,5,trackcellFD, (void*)&trackcellDFLT}, 
	{sizeof(ordercellT),6,1, ordercellFD, (void*)&ordercellDFLT},
	{sizeof(sampleT),   60+3,8,samplecellFD,(void*)&samplecellDFLT}
	}
#endif
;

/**************** PHYSICAL TRACKER DATA *****************/

#define MAXPATS 256
#define MAXCHNS 32
#define MAXSMPS 128
#define MAXORDER 256

typedef struct{
	trackcellT*d;
	int lgt;
}traxT;

typedef struct{
	int enabled;

	int panning;
}chaninfoT;

typedef struct{
	char name[32];
	char fname[64]; /* alloc? */
	int format;

	int globvol;
	int initspd;
	int inittmp;
	int mastvol;
	int stereo;
}songinfoT;

enum { FMAT_MOD, FMAT_S3M };

global traxT trax[MAXPATS][MAXCHNS];
global ordercellT order[MAXORDER];
global sampleT sample[MAXSMPS];  /* -> samples - type = samplecellT */
global chaninfoT chaninfo[MAXCHNS];
global songinfoT songinfo;

/******************** tABLE sTRUCTURES ****************/

#define MAXCOLS 16
columndef columns[MAXCOLS];
int tablesize;

/******************** EDiTOR iNTERFACE ****************/

#define attr_t int

enum {
	colTracks, colTracksBeat, colMarked,
	colColTitles, colMessages, colTitlebar};
global attr_t colAttr[6];
global WINDOW*w1,*w;

global struct{
	signed int
	mode,
	row0,col0,col9,
	row,col,fld,chr,
	oct,inst,
	brow0,brow1,bcol0,bcol1,
	
	playing,
	
	worthredraw,colschanged;
}es
#ifdef __MAIN_C
	={0,
	  0,0,0,
	  0,0,0,0,
	  2,0,
	  0,-1,0,-1,
	  
	  0,

	  0,0}
#endif
;

/******************** func prototypes *****************/

/*song.c*/
trackcellT*gettrack(int patt,int chan);
void packtrack(int patt,int chan);
int init_song();
void init_tracks();
void init_instruments();
int getnumch();
void pack_all();
int convertfx
    (uchar*fx,uchar*fxb,int targ);
    enum {TARG_MOD,TARG_S3M};
void convsmp0(char*d,int i,char*how);   /* rethink */
char*convsmp(sampleT*s,char*how);       /* rethink */
void write_song(char*fn);
void read_song(char*fn);
void write_sample(sampleT*s,char*fn);
void read_sample(sampleT*s,char*fn);
void delete_sample(sampleT*s);

/*editor.c*/
void setcol(int col,int ct,int patt,int chan);
void e_editor();
void e_resort_columns();
int e_getpattern();

/*player.c*/
void plr_init(int*argc,char**argv);
void plr_playsong();
void plr_stop();
void plr_setchan
     (int ch,sampleT*s,int note);
void plr_wake();
void plr_playpattern();
void plr_close();

/*ui.c*/
void drawtable
     (int col0,int row0,int y,int maxy);/* rethink */
void updateheadframe();                 /* rethink */
void drawheadframe(int col0);           /* rethink */
void e_mvcrs(int upd);                  /* rethink */
int e_getch(WINDOW*ww);
int e_getcmd();
void e_print(char*s);
void e_getstring(char*s,int lgt);
char*e_getfilename();

/****************** kEYBOARD COMMANDS ***********/

enum {
	cUnknown,
	cUp, cDown, cRight, cLeft,
	cUp16, cDown16, cTop, cBottom,
	cNextField, cPrevField, /* eiks cell? */
	cNextPattern,cPrevPattern,
	cNextOctave, cPrevOctave,
	cNextInst, cPrevInst,
	cJump,

	cBeginBlock,cEndBlock,cUnmarkBlock,
	cMarkEntireColumn,cMarkEntireTable,

	cCopyBlock,cPasteBlock,
	cInsOne,cDelOne,cInsBlock,cDelBlock,
	cProcess,
	cFileFuncs,

	cClearAll,cClearThis,cFillAll,cFillThis, /* kohta tarpeettomia*/
	cJumptoOrder,cResortColumns,cJumptoSamples, /* samoinkuin nm */

	cPlaySong,cPlayPattern,cPlayStop,

	cClearTracks,cClearInsts,cClearSong,

	cRedraw,cQuit
};
#define cNote 0x100
#define cNumeric 0x200
#define cChar 0x300
#define cInst 0x400

#define em_Notes 1
#define em_Numeric 2
#define em_Text 4

#undef global
