#include "oxt.h"

#ifdef HAVE_OSSDSP
#include <pthread.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#endif

typedef struct{
	sampleT*s;
	signed int off;
	signed int step;
	signed int vol;

	uint new;
	uint fx,fxb,note;
	signed int slideto;	
}p_chan;

struct{
	p_chan chan[MAXCHNS];
	
	int status;
	int outfreq;

	int order,row,tick,framelgt,speed;

	/***/
	
	int driver;
#ifdef HAVE_OSSDSP	
	int dev;
	pthread_t thr;int pid;
#endif
}plr;

enum { DRV_NOSOUND, DRV_OSSDSP };

enum { PL_OFF, PL_OFFREQ, PL_ROW, PL_PATT, PL_SONG, PL_TERM}; /* plr.status */
/*#define PL_HOLDIT 128*/

/************************/

void playframe();
void plrmain_wake(int sig);

void*plrmain(void*pim)
{
/*	signal(SIGUSR1,&plrmain_wake);*/
	while(plr.status!=PL_TERM){
		if((plr.status!=PL_OFF)&&(plr.status!=PL_OFFREQ))
			playframe();
		else
			{usleep(1000);}
		if(plr.status==PL_OFFREQ)plr.status=PL_OFF;
	}
	return NULL;
}

void plrmain_wake(int sig)
{
/*	plrmain(NULL);*/
}

void openaudio()
{
	int tmp;

	#ifdef HAVE_OSSDSP
	if(plr.driver==DRV_OSSDSP){

	plr.dev=open("/dev/dsp",O_WRONLY/*|O_NDELAY*/,0);	
	ioctl(plr.dev,SNDCTL_DSP_SPEED,&plr.outfreq);
	
	tmp=0;ioctl(plr.dev,SNDCTL_DSP_STEREO,&tmp);
	tmp=8;ioctl(plr.dev,SNDCTL_DSP_SAMPLESIZE,&tmp);
/*	tmp=(2<<16)|4096;ioctl(plr.dev,SNDCTL_DSP_SETFRAGMENT,&tmp);*/

	plr.status=PL_OFF;
	plr.pid=pthread_create(&plr.thr,NULL,&plrmain,NULL);
	
	}
	#endif
}

void setplaymode(int pl)
{
	int i;

	#ifdef HAVE_OSSDSP
	if(plr.driver==DRV_OSSDSP){

	if(pl==PL_OFF){
	plr.status=PL_OFFREQ;
	while(plr.status!=PL_OFF)usleep(10000);
	ioctl(plr.dev,SNDCTL_DSP_RESET,0);
	
	for(i=0;i<MAXCHNS;i++){
		plr.chan[i].step=0;
		plr.chan[i].new=0;
		plr.chan[i].s=NULL;
		plr.chan[i].fx=0;
		plr.chan[i].fxb=0;
	}
	}else{
		plr.status=pl;
/*		kill(plr.pid,SIGUSR1);*/
	}

	}
	#endif
}

void closeaudio()
{
	#ifdef HAVE_OSSDSP
	if(plr.driver==DRV_OSSDSP){
		setplaymode(PL_OFF);
/*		kill(plr.pid,SIGTERM);		*/
		close(plr.dev);
	}
	#endif
}

/************************/

int note2step(int note,sampleT*s);
void setbpm(int bpm);

void plr_init(int*argc,char**argv)
{
	int i;
	plr.driver=DRV_DEFAULT;
	plr.outfreq=FREQ_DEFAULT;

	for(i=1;i<*argc;i++)
	switch((int)argv[i][1]){
	case('d'):
		switch((int)argv[i][2]){
		case('n'):plr.driver=DRV_NOSOUND;break;
		#ifdef HAVE_OSSDSP
		case('d'):plr.driver=DRV_OSSDSP;break;
		#endif
		}break;
	case('f'):{
		int n=0;char*c=&argv[i][2];
		while(*c){if((*c>='0')&&(*c<='9'))n=(n*10)+(*c-'0');c++;}
		plr.outfreq=n;
	}break;
	}

	openaudio();
	setplaymode(PL_OFF);
}

void plr_playsong()
{
	setplaymode(PL_OFF);

	plr.order=plr.row=0;
	setbpm(songinfo.inittmp); /* elsewhere?*/
	plr.speed=plr.tick=songinfo.initspd; /* elsewhere?*/
	setplaymode(PL_SONG);
}

void plr_playpattern(int patt)
{
	setplaymode(PL_OFF);
	
	plr.row=0;plr.order=patt;
	setbpm(songinfo.inittmp); /* elsewhere?*/
	plr.speed=plr.tick=songinfo.initspd; /* elsewhere?*/
	setplaymode(PL_PATT);
}

void plr_stop()
{
	setplaymode(PL_OFF);
}

void plr_close()
{
	plr_stop();
	closeaudio();
}

void plr_setchan(int ch,sampleT*s,int note)
{
	if(s->data!=NULL){
	plr.chan[ch].s=s;
	plr.chan[ch].off=0;
	plr.chan[ch].step=note2step(note,s);
	plr.chan[ch].vol=s->volume;
	plr.chan[ch].new=1;
	}else plr.chan[ch].step=0;
}

void plr_wake()
{
	int oldch=0,allch=0,i;
	for(i=0;i<MAXCHNS;i++)
		if(plr.chan[i].step){
			allch++;
			if(!plr.chan[i].new)oldch++;
		}
	
	#ifdef HAVE_OSSDSP
	if(plr.driver==DRV_OSSDSP){
	
	if((allch)&&(!oldch)){
		plr.status=PL_OFFREQ;
		while(plr.status!=PL_OFF)usleep(10000);		
		ioctl(plr.dev,SNDCTL_DSP_RESET,0);
	}
	
	}
	#endif

	setplaymode(PL_ROW);
}

/************************/
/************************/
/************************/

#define per2step(val) (447408/(val)) /* ei toimi */
#define per2freq2(val) (447408/(val))

int addtoperiod(int step,int delta,int smpfrq)
{
	/* S1=cs/(cs/S+do) <=> */
	/* S1=c0s/(c0s/S+do/c1), c0c1=c */
	
	#define c0 (27*16)
	#define c1 (1024)

	int c0s,a;
	if(!step)return step;
	c0s=c0*smpfrq;	
	a=c0s/step+(delta*plr.outfreq)/c1;
	if(!a)return step;
	a=c0s/a;
	if((a<32)||(a>4096))return step;else return a;

	#undef c0
	#undef c1
}

/*447408=27*16384*/

int nper[12]={
27392,25856,24384,23040,21696,20480,19328,18240,17216,16256,15360,14496
};
int note2freq(int note)
{
	int oct=note/12,n=note%12;
   return per2step(nper[n]>>oct);
}
int note2step(int note,sampleT*s)
{
	return (s->freq*note2freq(note))/plr.outfreq;
}

signed int*mixbuf=NULL;
uchar*outbuf=NULL;
void setbpm(int bpm)
{
	plr.framelgt=plr.outfreq/(bpm*2/5);  /* settempo*/
	
	if(mixbuf!=NULL)free(mixbuf);
	if(outbuf!=NULL)free(outbuf);	
	mixbuf=malloc(plr.framelgt*sizeof(int)); /* *numchans*/
	outbuf=malloc(plr.framelgt*sizeof(char)); /* *samplesize*numchans*/
}

int p_ticker_0(trackcellT*n,p_chan*c)
{
	if(n->ins!=0xff){
	c->s=&sample[n->ins];
	if(c->s==NULL){c->s=&sample[0];}
	if((c->off>>8)>c->s->length)
		c->off=0;
		c->vol=c->s->volume;
	}

	if(n->vol!=0xff)
	c->vol=n->vol;

	{uchar fx=n->fx,fxb=n->fxb; int nxpat=0;
	if((fx!='0')&&(fxb==0)&&
	   ((fx==c->fx)||(c->fx='0')))
	   fxb=c->fxb;
	convertfx(&fx,&fxb,TARG_MOD);

	if(n->note!=0xff){

	int step=note2step(n->note,c->s);
/*	(c->s->freq*note2freq(n->note))/plr.outfreq;*/
	if(fx=='3')c->slideto=step;else c->step=step;
	
	c->off=0;
	c->note=n->note;
	}

	switch((int)fx){
		case('3'):
			if((c->step==0)&&(c->slideto==0))
				{fx='0';fxb=0;}
			/* varotoimenpide*/
			if(c->slideto==0)
				c->slideto=c->step;
			break;
		case('9'):
			c->off+=fxb<<16;
			break;
		case('B'):
			if(plr.status!=PL_PATT)plr.order=fxb;
			plr.row=0;
			break;
		case('C'):
			c->vol=fxb;
			break;
		case('D'):
			nxpat=(fxb<<8)|128;
			break;
		case('F'):if(fxb){
			if(fxb<=0x1f)plr.speed=plr.tick=fxb;
			else setbpm(n->fxb);}
			break;
	}
	c->fx=fx;
	c->fxb=fxb;	

	return nxpat;
}	}

void p_ticker_non0(p_chan*c)
{
	int fxb=c->fxb;
	if(fxb)switch((int)c->fx){

	case('0'):{ /* non-0-tick*/
		int foo=plr.tick%3,n=c->note;
		if(foo==1)n+=(fxb>>4);else	
		if(foo==2)n+=(fxb&0xf);
		c->step=(c->s->freq*note2freq(n))/plr.outfreq;
		}break;
	case('A'):if(plr.tick){
		int delta=0;
		if(fxb&0xf){
			if(fxb&0xf0)break;
			delta=-fxb;
		}else delta=fxb>>4;
		delta+=c->vol;
		if(delta<0)delta=0;else if(delta>64)delta=64;
		c->vol=delta;
		}break;
	case('1'):
		fxb=0-fxb;
	case('2'):
		if(plr.tick)if(c->step){
		c->step=addtoperiod(c->step,fxb<<2,c->s->freq);
		}break;
	case('3'):if(plr.tick){
		signed int step,diff;
		if(c->step<c->slideto)fxb=0-fxb; /*ylos*/
		step=addtoperiod(c->step,fxb<<2,c->s->freq);

		diff=c->slideto-step; /*matkaa ylos*/
		if(fxb>0)diff=0-diff; /*eiku alas*/
		if(diff<0)step=c->slideto;
		c->step=step;
		}break;
	}
}

void p_ticker()
{
	int patt;
	if(plr.status==PL_PATT)patt=plr.order;else patt=order[plr.order].pattno;

	if(plr.tick==plr.speed){

	{int i=0,nxpat=0;
	for(;i<MAXCHNS;i++){
	traxT*t=&trax[patt][i];
	if(t->d!=NULL)nxpat+=p_ticker_0(&t->d[plr.row],&plr.chan[i]);
	if(plr.row>=t->lgt-1)nxpat++;
	}
	
	plr.row++;
	if(nxpat){
		if(plr.status!=PL_PATT)plr.order++;
		plr.row=nxpat>>8;}
	if((plr.status==PL_SONG)&&(order[plr.order].pattno)==0xff)plr.order=0;
	}

	/* nextrow */

	plr.tick=0;
	};

	{int i=0;
	for(;i<MAXCHNS;i++){
	traxT*t=&trax[patt][i];
	if(t->d!=NULL)p_ticker_non0(&plr.chan[i]);
	}}
	
	plr.tick++;
}

void mix1chan(signed int*d,p_chan*c,int l0)
{
	signed char*s=c->s->data;
	int raja,looplgt;

	if(c->s->looping=='1'){
		raja=c->s->loopend;
		looplgt=(c->s->loopend-c->s->loopstart)<<8;
	}else{raja=c->s->length;looplgt=0;}
	raja<<=8;
		
	while((c->step)&&(l0>0)){
	int l=(raja-c->off)/c->step;
	if(l>l0)l=l0;
	l0-=l;

	{int coff=c->off,cstep=c->step,cvol=c->vol;
	for(;l>0;l--){
		*d+++=((s[(coff)>>8]*cvol)<<2);
		coff+=cstep;
	}
	c->off=coff;}
	
	if(l0){if(looplgt)c->off-=looplgt;else c->step=0;}
	}
}

void playframe()
{
	int i,numch=0;

	if((plr.status==PL_SONG)||(plr.status==PL_PATT))p_ticker();

	memset(mixbuf,0,plr.framelgt*sizeof(int));
	
	for(i=0;i<MAXCHNS;i++)if(chaninfo[i].enabled){
	p_chan*c=&plr.chan[i];
	c->new=0;
	if((c->step)&&(c->vol)/*&&(c->s!=NULL)&&(c->s->data!=NULL)*/){
		numch++;
		mix1chan(mixbuf,c,plr.framelgt);
		}
	}

/* todo: clipto_8bit_mono clipto_8bit_stereo clipto_16bit_mono
   clipto_16bit_stereo
*/
	{int l=plr.framelgt;uchar*d=outbuf;signed int*s=mixbuf;
	for(;l;l--){
		signed int c=128+(*s++>>9);
		if(c<0)c=0;else if(c>255)c=255;
		*d++=c;}
	}

	if((plr.status==PL_OFFREQ)||(plr.status==PL_OFF)){
		plr.status=PL_OFF;return;}

	#ifdef HAVE_OSSDSP
	write(plr.dev,outbuf,plr.framelgt);
	#endif
}
